import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * PocketServer manages and executes small, reusable, lightweight services on 
 * any type of device. It listens on a socket for instructions and can be used
 * locally or remotely.
 *
 * @author Christian Cantrell
 * @version $Id: PocketServer.java,v 1.11 2001/12/15 23:50:41 cantrell Exp $
 */
public class PocketServer
{
    // Properties loaded from the config file.
    private Properties props;

    // Path (w/i the jar file) to the config file.
    private static final String CONFIG_FILE = "config";

    // A simple logger.
    private PocketLogger logger;

    // This is where I keep all the services this server supports. They are
    // keyed off their names.
    private Hashtable services;

    // The "verbose" flag.
    private boolean verbose;

    // Whether the server is isolated or not. If the server is isolated,
    // it will ignore requests from any other hosts on the netword other
    // than the one the server is running on.
    private boolean isolated;

    public PocketServer()
        throws PocketException
    {
        props = new Properties();

        // Use the ClassLoader to load the Config file.
        try
        {
            InputStream configStream = 
                ClassLoader.getSystemResourceAsStream(CONFIG_FILE);
            props.load(configStream);
        }
        catch (Exception exception)
        {
            throw new PocketException("Problem loading the configuration " +
                                      "properties.", exception);
        }

        // Get the "verbose" property. If verbose is true, I periodically
        // output information about the server. If false, I'll output nothing. 
        String verStr = props.getProperty("verbose");
        verbose = (verStr == null) ? false : 
                                     Boolean.valueOf(verStr).booleanValue();
        logger = PocketLogger.getInstance();
        if (verbose) logger.log("PocketServer created.");

        // Get a ";"-delimited list of the PocketServices this server will
        // support.
        String allServices = props.getProperty("PocketServices");
        if (allServices == null)
        {
            throw new PocketException("No PocketServices were specified in " + 
                                      "your config file.");
        }
        StringTokenizer servicesTokens = new StringTokenizer(allServices,
                                                            ";");
        // Initialize the Hashtable we keep the services in.
        services = new Hashtable();
        while (servicesTokens.hasMoreTokens())
        {
            String className = null;
            try
            {
                className = servicesTokens.nextToken();
                Class serviceClass = Class.forName(className);
                PocketService pServ = 
                    (PocketService) serviceClass.newInstance();
                pServ.init(props);
                services.put(className, pServ);
                if (verbose)
                    logger.log(className + " successfully initialized.");
                
            }
            // So many different exceptions that can happen here that I'm just
            // going to catch a single Exception.
            catch (Exception exception)
            {
                throw new PocketException("Problem loading and initializing " +
                                          className + ".", exception);
            }
        }

        // Set the isolated property. If the server is configured to be
        // isolated, it will not service requests from any machine other
        // the one it is running on.
        String isolatedStr = props.getProperty("isolated");
        if (isolatedStr == null)
        {
            throw new PocketException("Cannot find \"isolated\" property in " +
                                      "config file.");
        } 
        isolatedStr = isolatedStr.toLowerCase();
        if (!isolatedStr.equals("true") && 
            !isolatedStr.equals("false"))
        {
            throw new PocketException("The \"isolated\" property must be " +
                                      "either \"true\" or \"false\".");
        }
        this.isolated = new Boolean(isolatedStr).booleanValue();
        if (verbose)
        {
            String v = (isolated) ? " " : " not ";
           logger.log("PocketServer is" + v + "isolated.");
        }

        // Start to set up the actual server.
        String portStr = props.getProperty("port");
        if (portStr == null)
        {
            throw new PocketException("Your config file does not specify a " +
                                      "port on which to listen.");
        }

        int port;
        try
        {
            port = Integer.parseInt(portStr);
        }
        catch (NumberFormatException exception)
        {
            throw new PocketException("The config property \"port\" must be " +
                                      "numeric.");
        }

        if (verbose) logger.log("PocketServer listening on port " + port + ".");

        ServerSocket server;
        try
        {
            server = new ServerSocket(port);
        }
        catch (IOException exception)
        {
            throw new PocketException("Problem opening a socket on port " +
                                      port + ".");
        }
        catch (SecurityException exception)
        {
            throw new PocketException("Problem opening a socket on port " +
                                      port + ". Try using an unprivileged " +
                                      "port above 1024.");
        }

        // Start the actual server.
        while (true)
        {
            try
            {
                new SocketHandler(server.accept());
            }
            catch (IOException exception)
            {
                throw new PocketException("An IOException was throws while " +
                                          "waiting for a connection.");
            }
        }
    }

    /**
     * An inner class for handling the socket connections. It extends Thread
     * because we want our server to be multi-threaded, meaning it can service
     * multiple requests simultaneously.
     */
    private class SocketHandler
        extends Thread
    {
        // The socket we are going to read from and write to. It's ok to
        // declare member variables at this point because a new instance of
        // a SocketHandler is made for every request.
        private Socket socket;
        private SocketHandler(Socket socket)
        {
            InetAddress address = socket.getInetAddress();

            // If the isolated property is set to true, make sure the request
            // has come from 127.0.0.1. 
            if (isolated)
            {
                if (!(address.toString().endsWith("127.0.0.1")))
                {
                    System.err.println("Rejecting request from [" +
                                       address + "].");
                    return;
                }
            } 

            this.socket = socket;
            if (verbose) logger.log("Connection from [" + address + "] accepted.");
            start();
        }

        public void run()
        {
            String fullRequest = null;
            PocketResponse response = null;
            try
            {
                // Get the InputStream from the Socket and read it in.
                InputStream in = socket.getInputStream();
                response = new PocketResponse(socket.getOutputStream());
                byte[] buffer = new byte[1024];
                ByteArrayOutputStream byteArrayOut =
                    new ByteArrayOutputStream();
                for (int cnt = in.read(buffer);
                     cnt != -1;
                     cnt = in.read(buffer))
                {
                    byteArrayOut.write(buffer, 0, cnt);
                    if (in.available() == 0)
                    {
                        break;
                    }
                }
                fullRequest = byteArrayOut.toString();
                if (verbose)
                {
                    logger.log("Full request: ");
                    logger.log(fullRequest);
                }
            }
            catch (Exception exception)
            {
                logger.log("Problem reading in a request.", exception);
                response.badRequest("An unknown error occurred while " +
                                    "reading in the request.", exception);
                return;
            }

            // Extract the portion of the request which pertains to service
            // instructions. Handle the StringIndexOutOfBoundsException as an
            // error.
            String service = null;
            PocketRequest request = null;
            try
            {
                String serviceRequest =
                    fullRequest.substring(fullRequest.indexOf("/"),
                                          fullRequest.indexOf(" ",
                                          fullRequest.indexOf("/")));

                // Check to see if the server should exit.
                if (serviceRequest.indexOf("?") == -1 && 
                    serviceRequest.endsWith("KillServer"))
                {
                    System.exit(0);
                }

                // Now parse the service instructions.
                service = serviceRequest.substring(serviceRequest.indexOf("/") +
                                                   1,
                                                   serviceRequest.indexOf("?"));

                if (verbose) logger.log("A request was made for service [" +
                                        service + "] with the following " +
                                        "arguments: ");
                String allArgs =
                    serviceRequest.substring(serviceRequest.indexOf("?") + 1,
                                             serviceRequest.length());
                StringTokenizer argTokens = new StringTokenizer(allArgs, "&");
                request = new PocketRequest();
                while (argTokens.hasMoreTokens())
                {
                    String arg = argTokens.nextToken();
                    String name = arg.substring(0,arg.indexOf("="));
                    String val = arg.substring(arg.indexOf("=") + 1,
                                               arg.length());
                    val = URLDecoder.decode(val);
                    if (verbose)
                    {
                        logger.log(name + "=" + val);
                    }
                    request.put(name, val);
                }
            }
            catch (StringIndexOutOfBoundsException exception)
            {
                logger.log("A malformed request was made. Requests should " +
                           "look like: http://<ip>:<port>/<PocketService>?" +
                           "<argName>=<argValue>&<argName>=<argValue>");
                response.badRequest("Malformed request.", exception);
                return;
            }

            // Now get the requested PocketService from the Hashtable of
            // PocketServices and call service() on it.
            PocketService pServ = (PocketService) services.get(service);
            if (pServ == null)
            {
                logger.log("Error: A request was made for the unknown " +
                           "PocketService [" + service + "]");
                response.notFound("Resource not found. [" + service +
                                  "] does not exist on this " + "server.",null);
                return;
            }
            try
            {
                pServ.service(request,response);
            }
            catch (PocketServiceException exception)
            {
                logger.log("The PocketService " + service + " threw a " +
                           "PocketServiceException.", exception);
                response.internalServerError("Internal Server Error. ",
                                             exception);
                return;
            }
        }
    }

    public static void main(String[] argv)
    {
        try
        {
            new PocketServer();
        }
        catch (PocketException exception)
        {
            System.err.println("Server Failed:");
            exception.printStackTrace();
            System.exit(-1);
        }
    }
}
